/*  
    This file is part of the KDE project
    Copyright (C) 1998 Ulrich Clanget (clanget@studcs.uni-sb.de)
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
    
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
    
    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
    
    
    Put this file in your boot startup file. Sets the old mixer configuration, saved after
    closing ossmixer. If your mixer devices differs from /dev/mixer, there is an option
    to change the default mixer device.
    
    mixerconfig [Mixer Device]
    
    e.g: mixerconfig /dev/mixer1
    
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

// oss include

#include <sys/soundcard.h>
#include <sys/ioctl.h>

// device management include

#include <fcntl.h>

// consts

#define CHANNEL_SENSITIVITY    10
#define DEVICE_STEP            50

typedef struct SDeviceInfo
{
 int VolumeLeft;
 int VolumeRight;
 int Volume;
 int ChannelVal;
 bool Stereo;
 bool RecordSrc;
 bool Record;
 bool Muting;
 bool Init;
 bool Show;
};

typedef struct SMixerDevice
{
 int Handle;
 int DeviceMask;
 int RecMask;
 int StereoDeviceMask;
 int RecSrcMask;
 int Capabilities;
 bool ExclusiveInput;
};

// global variables

SDeviceInfo Device[SOUND_MIXER_NRDEVICES];
SMixerDevice MixerDevice;
mixer_info MixerInfo;

// display errormessages and exit

void Error(const char* Error)
{
  printf("ossmixer error: ");
  printf(Error);
  printf("\n");
  exit(-1);
}

bool findDevice(char Devicename)
{
  if (MixerDevice.DeviceMask & (1 << Devicename)) return true; else return false;
}

void toggleRecButtons()
{
  for (int i = 0; i < SOUND_MIXER_NRDEVICES; i++)
    {
      if (Device[i].Record)
	if ((1 << i) & MixerDevice.RecSrcMask) 
	  {
	    Device[i].RecordSrc = true;
	  }
	else
	  {
	    Device[i].RecordSrc = false;
	  } 
    }
}

void loadConfiguration()
{  
  char* home = getenv("HOME");
  char* configFileName = "/.OSSMixer.config";
    
  char string[strlen(home) + strlen(configFileName) + 1];
  
  strncpy(string, home, strlen(home));
  strcpy(string + strlen(home), configFileName);

  printf("ossmixer: Loading configuration file\n");

  FILE *File;
  
  File = fopen(string, "rb");
  
  if (!File)
    Error("can't load config file");
  else
    { 
      fread(Device, sizeof(SDeviceInfo), SOUND_MIXER_NRDEVICES, File);
      fclose(File);
    }
}

void recordChanged(int Devname, bool State)
{
  if (MixerDevice.ExclusiveInput)
    {
      MixerDevice.RecSrcMask = 1 << Devname;
    } 
  else 
    {
      // create record src mask
      MixerDevice.RecSrcMask = 0;
      int BoolInt;
      //      Device[Devname].RecordSrc = State; BUG
      for (int i = 0; i < SOUND_MIXER_NRDEVICES; i++)
	{
	  if (Device[i].RecordSrc) BoolInt = 1; else BoolInt = 0;
	  MixerDevice.RecSrcMask = MixerDevice.RecSrcMask | ( BoolInt << i);
	}
    }  
  
  if (ioctl(MixerDevice.Handle, SOUND_MIXER_WRITE_RECSRC, &MixerDevice.RecSrcMask) == -1)
    Error("Can't change Record-Source");

  toggleRecButtons();
}

void channelChanged(int Devname, int Value)
{ 
  int Vol = Device[Devname].VolumeLeft | (Device[Devname].VolumeRight << 8);
  if (Device[Devname].Muting) Vol = 0;
  
  if (ioctl(MixerDevice.Handle, MIXER_WRITE(Devname), &Vol) == -1)
    Error("Can't change Volume");
}

void setVolume(int Devname, int Volume)
{
  Device[Devname].Volume = Volume;
  channelChanged(Devname, Device[Devname].ChannelVal);
}



// main

void main(int argc, char** argv)
{
  // default mixer-device
  char* mixerDevice  = "/dev/mixer";
  
  if (argc > 1) mixerDevice = argv[argc - 1];

  // open mixer device
  MixerDevice.Handle = open(mixerDevice, O_RDONLY);
  
  if (MixerDevice.Handle == -1) 
    Error("Can't open Mixer Device");

  // get mixer info
  if (ioctl(MixerDevice.Handle, SOUND_MIXER_INFO, &MixerInfo) == -1)
    Error("Can't get Mixerinfo");
  
  // get MIXER_READ_DEVMASK
  if (ioctl(MixerDevice.Handle, SOUND_MIXER_READ_DEVMASK, &MixerDevice.DeviceMask) == -1) 
    Error("Can't get supported Devices");
  
  // get MIXER_READ_STEREODEVS
  if (ioctl(MixerDevice.Handle, SOUND_MIXER_READ_STEREODEVS, &MixerDevice.StereoDeviceMask) == -1) 
    Error("Can't get supported Stereo Devices");
  
  // get MIXER_READ_RECMASK
  if (ioctl(MixerDevice.Handle, SOUND_MIXER_READ_RECMASK, &MixerDevice.RecMask) == -1) 
    Error("Can't get supported Recorddevices");
  
  // get MIXER_READ_RECSRC
  if (ioctl(MixerDevice.Handle, SOUND_MIXER_READ_RECSRC, &MixerDevice.RecSrcMask) == -1) 
    Error("Can't get Recordsource");
  
  // get MIXER_READ_CAPS
  if (ioctl(MixerDevice.Handle, SOUND_MIXER_READ_CAPS, &MixerDevice.Capabilities) == -1) 
    Error("Can't get Mixer capabilities");
  
  // only one recording channel at the same time?
  if ((1 << SOUND_CAP_EXCL_INPUT) & MixerDevice.Capabilities) MixerDevice.ExclusiveInput = true; else
    MixerDevice.ExclusiveInput = false;
  
  // set device properties
  for (int i = 0; i < SOUND_MIXER_NRDEVICES; i++)
    {
      // device support stereo?
      if ((1 << i) & MixerDevice.StereoDeviceMask) Device[i].Stereo = true; else Device[i].Stereo = false;
      
      // device support record?
      if ((1 << i) & MixerDevice.RecMask) Device[i].Record = true; else Device[i].Record = false;
      
      // device is recordsrc? 
      if ((1 << i) & MixerDevice.RecSrcMask) Device[i].RecordSrc = true; else Device[i].RecordSrc = false;
      
      Device[i].ChannelVal = 0;
    } 
  

  // load configuration file
  loadConfiguration();

  printf("ossmixer: Setting configuration\n");

  // configure if configuration file is loaded
  
  for (int i = 0; i < SOUND_MIXER_NRDEVICES; i++)
    if (findDevice(i))
      {
	setVolume( i, Device[i].Volume );
	recordChanged( i, Device[0].RecordSrc );
      }
}

