/*  
    This file is part of the KDE project
    Copyright (C) 1998 Ulrich Clanget (clanget@studcs.uni-sb.de)
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
    
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
    
    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.

*/

#include "ossmixer.moc"
#include "icons.h"
#include <kmsgbox.h>

// global variables

SDeviceInfo Device[SOUND_MIXER_NRDEVICES];
SMixerDevice MixerDevice;
mixer_info MixerInfo;

// display error message and exit

void Error(const char* Error)
{
  KMsgBox::message(0, "OSSMixer Error Message", Error, KMsgBox::STOP);
  exit(-1);
}

uchar IconID[SOUND_MIXER_NRDEVICES] = { 4, 0, 10, 9, 11, 8, 2, 5, 1, 4, 11, 4, 4, 4, 2, 2, 2 };

char* DeviceNameString[SOUND_MIXER_NRDEVICES] = { "Master Volume", "Bass", "Treble", "Synthesizer (FM / Wavetable)",
                                                  "PCM" , "Speaker", "Line In", "Microphone", "Compact Disc",
                                                  "Recording Monitor", "Alternate PCM", "Recording level", "Input gain",
                                                  "Output gain", "AUX 1", "AUX 2", "Line In 2" }; 

OMixer::OMixer(const char* MixerDeviceName, QWidget* parent, const char* name)
{
  setFrameStyle(QFrame::Panel | QFrame::Raised);
  
  // open mixer device
  MixerDevice.Handle = open(MixerDeviceName, O_RDONLY);
  if (MixerDevice.Handle == -1) 
    Error("Can't open Mixer Device");
  
  // get mixer info
  if (ioctl(MixerDevice.Handle, SOUND_MIXER_INFO, &MixerInfo) == -1)
    Error("Can't get Mixerinfo");
  
  // get MIXER_READ_DEVMASK
  if (ioctl(MixerDevice.Handle, SOUND_MIXER_READ_DEVMASK, &MixerDevice.DeviceMask) == -1) 
    Error("Can't get supported Devices");
  
  // get MIXER_READ_STEREODEVS
  if (ioctl(MixerDevice.Handle, SOUND_MIXER_READ_STEREODEVS, &MixerDevice.StereoDeviceMask) == -1) 
    Error("Can't get supported Stereo Devices");
  
  // get MIXER_READ_RECMASK
  if (ioctl(MixerDevice.Handle, SOUND_MIXER_READ_RECMASK, &MixerDevice.RecMask) == -1) 
    Error("Can't get supported Recorddevices");
  
  // get MIXER_READ_RECSRC
  if (ioctl(MixerDevice.Handle, SOUND_MIXER_READ_RECSRC, &MixerDevice.RecSrcMask) == -1) 
    Error("Can't get Recordsource");
  
  // get MIXER_READ_CAPS
  if (ioctl(MixerDevice.Handle, SOUND_MIXER_READ_CAPS, &MixerDevice.Capabilities) == -1) 
    Error("Can't get Mixer capabilities");
  
  // only one recording channel at the same time?
  if ((1 << SOUND_CAP_EXCL_INPUT) & MixerDevice.Capabilities) MixerDevice.ExclusiveInput = true; else
    MixerDevice.ExclusiveInput = false;
  
  // set device properties
  for (int i = 0; i < SOUND_MIXER_NRDEVICES; i++)
    {
      // device support stereo?
      if ((1 << i) & MixerDevice.StereoDeviceMask) Device[i].Stereo = true; else Device[i].Stereo = false;
      
      // device support record?
      if ((1 << i) & MixerDevice.RecMask) Device[i].Record = true; else Device[i].Record = false;
      
      // device is recordsrc? 
      if ((1 << i) & MixerDevice.RecSrcMask) Device[i].RecordSrc = true; else Device[i].RecordSrc = false;
      
      // disable muting
      Device[i].Muting = false;
      
      // no device initialized
      Device[i].Init = false;
      
      // get current volume
      if (findDevice(i)) getVolume(i);
      
      // show all devices at begin
      Device[i].Show = true;
      
      // channelval
      Device[i].ChannelVal = 0;
    } 
  
  bool ConfigFile = false;
  if (findConfigurationFile()) 
    {
      ConfigFile = true;
      loadConfiguration();
    } 
  
  // init all supported devices
  
  for (int f = 0; f < SOUND_MIXER_NRDEVICES; f++)
    {
      if (findDevice(f))
	{
	  initDevice(f, IconID[f]);
	}
    }
  
  // display logoframe
  
  Frame = new QFrame(this);
  CHECK_PTR(Frame);
  QPixmap Pix;
  Pix.loadFromData(MixerIcons[3].data, MixerIcons[3].size); 
  Frame->setFrameStyle(QFrame::Panel | QFrame::Raised);
  Frame->resize(Pix.width(), Pix.height());
  Frame->setBackgroundPixmap(Pix);
  Frame->show();
  
  setCaption("Open Sound System Mixer");
  
  // configure
  
  if (ConfigFile) 
    {
      for (int i = 0; i < SOUND_MIXER_NRDEVICES; i++)
	if (Device[i].Init)
	  {
	    DeviceWidget[i]->setVolume( Device[i].Volume );
	  }
      DeviceWidget[0]->recordChanged( Device[0].RecordSrc );
    } 
  
  // init config menu
  
  config = new QPopupMenu;
  CHECK_PTR( config );
  
  config->setCheckable( TRUE );
  
  for (int i = 0; i < SOUND_MIXER_NRDEVICES; i++)
    {
      Pix.loadFromData(MixerIcons[IconID[i]].data, MixerIcons[IconID[i]].size);  
      if (Device[i].Init) 
	{
	  config->insertItem( Pix , i ); 
	  if (Device[i].Show) config->setItemChecked( i, true ); 
	  else config->setItemChecked( i, false );
	}
    }
  
  connect(config, SIGNAL(activated(int)), SLOT(configDevices(int)));
  
  // init menu bar
  
  file = new QPopupMenu;
  CHECK_PTR( file );
  file->insertItem( "MixerInfo", this, SLOT(displayMixerInfo()) );
  file->insertItem( "Exit", this, SLOT(exit()));
  
  help = new QPopupMenu;
  CHECK_PTR( help );
  help->insertItem( "About", this, SLOT(about()) );
  
  menu = new QMenuBar( this );
  CHECK_PTR(menu);
  menu->insertItem( "&Mixer", file);
  menu->insertItem( "&Devices", config);
  menu->insertItem( "&Help", help);
  
  move(0,0);
  
  showDevices();
  
  show();
}

OMixer::~OMixer()
{ 
}

void OMixer::showDevices()
{
  int Count = 10;
  
  for (int i = 0; i < SOUND_MIXER_NRDEVICES; i++)
    {
      if (Device[i].Init)
	if (Device[i].Show)
	  {
	    DeviceWidget[i]->move(Count , 40);
	    Count = Count + DEVICE_STEP;
	    DeviceWidget[i]->show();
	  }
	else DeviceWidget[i]->hide(); 
    } 
  
  Frame->move(Count + 20, 40);
  setMaximumSize(Count + 20 + Frame->width() + 20, 270);
  resize(Count + 20 + Frame->width() + 20, 270);
}

bool OMixer::findDevice(char Devicename)
{
  if (MixerDevice.DeviceMask & (1 << Devicename)) return true; else return false;
}

void OMixer::getVolume(int Devicename)
{
  int Volume;
  if (ioctl(MixerDevice.Handle, MIXER_READ(Devicename), &Volume) == -1)
  Error("Can't get Device Volume");
  
  int VolumeLeft = Volume & 0xff; // get low byte
  int VolumeRight = Volume >> 8;  // get high byte
  
  Device[Devicename].VolumeLeft = VolumeLeft;
  Device[Devicename].VolumeRight = VolumeRight;
  
  Device[Devicename].Volume = QMAX(VolumeLeft, VolumeRight);
  if (VolumeLeft < VolumeRight) Device[Devicename].ChannelVal = (100 - VolumeLeft) / CHANNEL_SENSITIVITY;
  if (VolumeLeft > VolumeRight) Device[Devicename].ChannelVal = -(100 - VolumeRight) / CHANNEL_SENSITIVITY;
  if (VolumeLeft == VolumeRight) Device[Devicename].ChannelVal = 0;
}

void OMixer::initDevice(int Devicename, int PicIndex)
{
  DeviceWidget[Devicename] = new ODeviceWidget(Devicename, PicIndex, this);
  CHECK_PTR(DeviceWidget[Devicename]);
  connect(DeviceWidget[Devicename], SIGNAL(toggleRecButtons()), SLOT(toggleRecButtons()));
  Device[Devicename].Init = true;
}

void OMixer::toggleRecButtons()
{
  for (int i = 0; i < SOUND_MIXER_NRDEVICES; i++)
    {
      if (Device[i].Record)
	if ((1 << i) & MixerDevice.RecSrcMask) 
	  {
	    DeviceWidget[i]->RecButton->setOn(true);
	    Device[i].RecordSrc = true;
	  }
	else
	  {
	    DeviceWidget[i]->RecButton->setOn(false);
	    Device[i].RecordSrc = false;
	  } 
    }
}

void OMixer::configDevices(int MenuIndex)
{
  Device[MenuIndex].Show = !Device[MenuIndex].Show;
  config->setItemChecked(MenuIndex,Device[MenuIndex].Show);
  showDevices();
}

void OMixer::displayMixerInfo()
{
  QMessageBox msg(0, "Mixerinfo");
  QPixmap pix;
  pix.loadFromData(MixerIcons[6].data, MixerIcons[6].size);
  msg.setIconPixmap(pix);
  msg.setCaption("Mixer Info");
  QString tmp(MixerInfo.id);
  tmp += "\n";
  tmp += MixerInfo.name;
  msg.setText(tmp);
  msg.exec();
}

void OMixer::about()
{
  KMsgBox::message(0, "About OSS Mixer", "Open Sound System (OSS) Mixer Final \n(c) Apr 1998 Ulrich Clanget\nE-mail: clanget@studcs.uni-sb.de\nThis is free software and comes with no warranty\nSee the GNU General Public License for details.", KMsgBox::INFORMATION); 
}

void OMixer::saveConfiguration()
{
  QString ConfigFileName(getenv("HOME"));
  ConfigFileName.append("/.OSSMixer.config");
  
  FILE *File;
  
  File = fopen(ConfigFileName, "wb");
  
  if (!File)
    Error("Can't write config file");
  else
    { 
      fwrite(Device, sizeof(SDeviceInfo), SOUND_MIXER_NRDEVICES, File);
      fclose(File);
    } 
}

bool OMixer::findConfigurationFile()
{
  QString ConfigFileName(getenv("HOME"));
  ConfigFileName.append("/.OSSMixer.config");
  QFile File(ConfigFileName);
  return File.exists(); 
}

void OMixer::loadConfiguration()
{
  if (findConfigurationFile())
    {
      QString ConfigFileName(getenv("HOME"));
      ConfigFileName.append("/.OSSMixer.config");
      
      FILE *File;
      
      File = fopen(ConfigFileName, "rb");
      
      if (!File)
	Error("Can't load config file");
      else
	{ 
	  fread(Device, sizeof(SDeviceInfo), SOUND_MIXER_NRDEVICES, File);
	  fclose(File);
	} 
    }
}

void OMixer::exit()
{
  saveConfiguration();
  qApp->quit();
}

void OMixer::closeEvent( QCloseEvent* )
{
  saveConfiguration();
  qApp->quit();
}


// device widget

ODeviceWidget::ODeviceWidget(int Devicename, int PicIndex, QWidget* parent, const char* name)
  : QFrame(parent, name)
{
  Devname = Devicename;
  setGeometry(0, 40, 44, 210);
  
  // draw range marks
  
  QPixmap Buffer(width(), height());
  Buffer.fill(lightGray);
  QPainter Paint;
  Paint.begin(&Buffer);
  
  for (int i = 65; i < height(); i = i +20)
    {
      Paint.setPen(darkGray);
      Paint.drawLine(5, i, width() - 5, i);
      Paint.setPen(white);
      Paint.drawLine(5, i+1, width() - 5, i+1);
    }
  
  Paint.end();
  
  setBackgroundPixmap(Buffer);
  
  setFrameStyle(QFrame::Panel | QFrame::Raised);
  QPixmap Pix;
  Pix.loadFromData(MixerIcons[PicIndex].data, MixerIcons[PicIndex].size);
  DeviceFrame = new QFrame(this);
  CHECK_PTR(DeviceFrame);
  DeviceFrame->setGeometry(2, 2, Pix.width(), Pix.height());
  DeviceFrame->setFrameStyle(QFrame::Panel | QFrame::Raised);
  DeviceFrame->setBackgroundPixmap(Pix);
  DeviceFrame->show();   
  
  QToolTip::add( DeviceFrame, DeviceNameString[Devname] );
  
  VolumeSlider = new OVolumeSlider(this);
  CHECK_PTR(VolumeSlider);
  VolumeSlider->setGeometry(13, 65, 20, 120);
  VolumeSlider->setValue(100 - Device[Devicename].Volume);
  VolumeSlider->show();  
  
  connect(VolumeSlider, SIGNAL(valueChanged(int)), SLOT(volumeChanged(int)));
  
  ChannelSlider = new QScrollBar(-CHANNEL_SENSITIVITY, CHANNEL_SENSITIVITY, 
				 1, 1, Device[Devicename].ChannelVal, QScrollBar::Horizontal, this);
  CHECK_PTR(ChannelSlider);
  ChannelSlider->setGeometry(2, 192, width()-4, 12);
  if (Device[Devname].Stereo) ChannelSlider->show(); else ChannelSlider->hide();
  
  connect(ChannelSlider, SIGNAL(valueChanged(int)), SLOT(channelChanged(int)));
  
  RecButton = new QPushButton(this);
  CHECK_PTR(RecButton);
  RecButton->setGeometry(6,45,15,15);
  RecButton->setToggleButton(true);
  Pix.loadFromData(MixerIcons[7].data, MixerIcons[7].size);
  RecButton->setPixmap(Pix);
  if (Device[Devname].Record) RecButton->show(); else RecButton->hide();
  RecButton->setOn(Device[Devname].RecordSrc);
  
  connect(RecButton, SIGNAL(toggled(bool)), SLOT(recordChanged(bool)));
  
  MutingButton = new QPushButton(this);
  CHECK_PTR(MutingButton);
  MutingButton->setGeometry(23,45,15,15);
  MutingButton->setToggleButton(true);
  Pix.loadFromData(MixerIcons[12].data, MixerIcons[12].size);
  MutingButton->setPixmap(Pix);
  if (!Device[Devname].Record) MutingButton->move(15,45);
  MutingButton->show();
  MutingButton->setOn(Device[Devname].Muting);
  
  connect(MutingButton, SIGNAL(clicked()), SLOT(mutingChanged()));
}

ODeviceWidget::~ODeviceWidget()
{
}

void ODeviceWidget::volumeChanged(int Volume)
{
 setVolume(100 - Volume);
}

void ODeviceWidget::mutingChanged()
{
  Device[Devname].Muting = !Device[Devname].Muting;
  setVolume(Device[Devname].Volume);
}

void ODeviceWidget::recordChanged(bool State)
{
  if (MixerDevice.ExclusiveInput)
    {
      MixerDevice.RecSrcMask = 1 << Devname;
    } 
  else 
    {
      // create record src mask
      MixerDevice.RecSrcMask = 0;
      int BoolInt;
      Device[Devname].RecordSrc = State;
      for (int i = 0; i < SOUND_MIXER_NRDEVICES; i++)
	{
	  if (Device[i].RecordSrc) BoolInt = 1; else BoolInt = 0;
	  MixerDevice.RecSrcMask = MixerDevice.RecSrcMask | ( BoolInt << i);
	}
    }  
  
  if (ioctl(MixerDevice.Handle, SOUND_MIXER_WRITE_RECSRC, &MixerDevice.RecSrcMask) == -1)
    Error("Can't change Record-Source");
  
  emit toggleRecButtons();
}

void ODeviceWidget::channelChanged(int Value)
{
  Device[Devname].ChannelVal = Value;
  if (Value < 0) 
    {
      Device[Devname].VolumeLeft = Device[Devname].Volume;
      Device[Devname].VolumeRight = Device[Devname].Volume + Value*Device[Devname].Volume/CHANNEL_SENSITIVITY;
    }
  if (Value > 0)
    {
      Device[Devname].VolumeRight = Device[Devname].Volume;
      Device[Devname].VolumeLeft = Device[Devname].Volume - Value*Device[Devname].Volume/CHANNEL_SENSITIVITY;
    } 
  if (Value == 0) 
    {
      Device[Devname].VolumeRight = Device[Devname].Volume;
      Device[Devname].VolumeLeft  = Device[Devname].Volume;
    }
  
  int Vol = Device[Devname].VolumeLeft | (Device[Devname].VolumeRight << 8);
  if (Device[Devname].Muting) Vol = 0;
  
  if (ioctl(MixerDevice.Handle, MIXER_WRITE(Devname), &Vol) == -1)
    Error("Can't change Volume");
}

void ODeviceWidget::setVolume(int Volume)
{
  Device[Devname].Volume = Volume;
  channelChanged(ChannelSlider->value());
}


// volume slider

OVolumeSlider::OVolumeSlider(QWidget* parent, const char* name)
  : QSlider(0, 100, 1, 0 , QSlider::Vertical, parent, name)
{
}

void OVolumeSlider::paintSlider(QPainter *Paint, const QRect& Range)
{
  Paint->setBrush(lightGray);
  Paint->setPen(darkGray);
  Paint->drawRect(Range);
  Paint->setPen(white);
  Paint->drawLine(Range.x(), Range.y(), Range.right(), Range.y());
  Paint->drawLine(Range.x(), Range.y(), Range.x(), Range.bottom());
}





